/* Copyright (C) 1994 Queen Mary and Westfield College, University of London

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   A copy of the GNU General Public License may be found in
   doc/COPYING in the source tree of this programm.

   This file was written by: Franz Gerschwiler.(goblin@dcs.qmw.ac.uk)

 ****

   modified by: p.hlstede	(peter@calvados.chemie.uni-bremen.de)

   - the code was reorganized and ansified
   - many resource settings were moved to the app-defaults file
   - added filename dependent wordwrap switching
   - a '-' as filename means that gloe reads from stdin
   - a normal file selection box is now use for 'save as' operations

							06.09.1997
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>

#include <Xm/Xm.h>
#include <Xm/FileSB.h>
#include <Xm/Text.h>

#include "sys_consts.h"
#include "globals.h"
#include "utils.h"
#include "undo.h"
#include "callbacks.h"
#include "fileEvents.h"
#include "optionEvents.h"
#include "dstring.h"


/***  PROCEDURE DECLARATIONS  ***/
static void    set_filename(char *filename);
static void    gui_createFileDialog(Widget parent, FileMode mode);
static void    gui_cancelBrowseCB(Widget w, XtPointer client_data, XtPointer call_data);
static void    gui_browseFileCB(Widget w, XtPointer client_data, XtPointer call_data);
static void    AutoWW(char *name);
static void    gui_saveFileCB(Widget w, XtPointer client_data, XtPointer call_data);
static Boolean gui_save(char *filename, FileMode mode);
static Boolean savefile(char *filename, FileMode mode);



void AutoSaveFunc(XtPointer client_data, XtIntervalId *id)
{
   DString  filename;
   char    *home, *name;

   set_cursor(toplevel, TimeCursor);
   set_cursor(text_canvas, TimeCursor);
   set_status_message("Please wait: Autosaving");

   name = strrchr(FileName, '/');
   if (name) name++;
   else      name = FileName;

   home = getenv("HOME");
   if (!home) home = "";

   DStringCompose(&filename, home, GloePath, "/", name, NULL);
   if (!savefile(filename.str, SAVE_ALL))
      gui_error("Sorry unable to perform autosave");
   DStringFree(&filename);

   set_status_message("");
   unset_cursor(toplevel);
   unset_cursor(text_canvas);

   if (AutoSave)
      XtAppAddTimeOut(app_context, AutoSaveTime*1000, AutoSaveFunc, client_data);
}


void  gui_browseCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   gui_createFileDialog(toplevel, (int)client_data);
}


/* Reads in the file */
Boolean  loadfile(char *filename, FileMode mode)
{
   int      stdin_flag = !strcmp(filename, "-");
   struct stat stat_val;
   char    *filecontents;
   FILE    *fpointer;
   int      akt_len;
   int      f_length;

   if (stdin_flag || !stat(filename, &stat_val))
   {
      /* read from STDIN if the filename is '-' */
      if (stdin_flag)
      {
         fpointer = stdin;
         f_length = 1024;	/* length of one block to read */
      }
      else
      {
         f_length = stat_val.st_size;
         fpointer = fopen(filename, "r");
      }
      filecontents = NULL;
      akt_len = 0;
      if (fpointer != NULL)
      {
	 /* read in the file in blocks of size f_length. */
	 /* only one read is necessary for normal files. */
	 while (f_length && !feof(fpointer))
	 {
	    filecontents = XtRealloc(filecontents, (unsigned)(akt_len + f_length + 1));
	    akt_len += fread(filecontents+akt_len, sizeof(char), f_length, fpointer);
	    filecontents[akt_len] = '\0';
	 }
	 /* don't close STDIN */
	 if (!stdin_flag)
	    fclose(fpointer);

	 /* Put data  read in into the text widget wiping out old text */
	 if (mode == OPENIT)
	 {
	    undo_redo_action = True;
	    if (filecontents)
	       XmTextSetString(text_canvas, filecontents);
	    else
	       XmTextSetString(text_canvas, "");
	    reset_undo_redo();
	    AutoWW(filename);
	    set_icon_name(filename);
	 }
	 else if (mode == APPENDIT)
	 {
	    /* Insert data read in into current text */
	    if (filecontents)
	       XmTextInsert(text_canvas, XmTextGetLastPosition(text_canvas), filecontents);
	 }
	 else
	 {
	    /* Insert data read in into current text */
	    if (filecontents)
	       XmTextInsert(text_canvas, XmTextGetInsertionPosition(text_canvas), filecontents);
	 }
	 XtFree(filecontents);
	 return(True);
      }
      else
	 return(False);
   }
   else
   {
      if (errno == ENOENT)
      {
         AutoWW(filename);
         if (mode == OPENIT)
	    set_icon_name(filename);
      }
      return False;
   }
}




static void set_filename(char *filename)
{
   XtFree(FileName);
   FileName = filename;
   set_title_string();
}

static void  gui_createFileDialog(Widget parent, FileMode mode)
{
   Widget  fileD;
   Arg     args[5];
   int     n = 0;
   char   *w_name = NULL;

   switch(mode)
   {
      case APPENDIT:
      case OPENIT :
         w_name = "openFileD";
         break;

      case INSERTIT :
         w_name = "insertFileD";
         break;

      case SAVE_AS:
      case SAVE_ALL:
         w_name = "saveFileD";
         break;

      case SAVE_SEL:
         w_name = "saveSelFileD";
         break;
   }

   XtSetArg(args[n], XmNdeleteResponse, XmDESTROY); n++;
   fileD = XmCreateFileSelectionDialog(parent, w_name, args, n);

   if ((mode == OPENIT) || (mode == INSERTIT))
      XtAddCallback(fileD, XmNokCallback, gui_browseFileCB, (XtPointer)mode);
   else
      XtAddCallback(fileD, XmNokCallback, gui_saveFileCB, (XtPointer)mode);

   XtAddCallback(fileD, XmNcancelCallback, gui_cancelBrowseCB, (XtPointer)mode);

   XtUnmanageChild(XmFileSelectionBoxGetChild(fileD, XmDIALOG_HELP_BUTTON));

   XtManageChild(fileD);
}

static void  gui_cancelBrowseCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   XtDestroyWidget(w);
}

static void  gui_browseFileCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   XmFileSelectionBoxCallbackStruct *cbd =
		(XmFileSelectionBoxCallbackStruct *)call_data;
   FileMode mode = (FileMode)client_data;
   char    *filename = NULL;
   char     answer;

   /* Get the filename from the file selection box */
   XmStringGetLtoR(cbd->value, CHARSET, &filename);

   /* Make sure the user doesn't lose data accidentally */
   if (mode == OPENIT)
   {
      if (changes_counter)
	 answer = gui_query("Unsaved Text. Are you sure about this?");
      else
         answer = YES;

      if (answer == SAVE_THEN_DO)
         gui_saveCB(w, NULL, NULL);

      if ((answer == YES) || (answer == SAVE_THEN_DO))
      {
         set_status_message("Attempting to open file");
	 set_cursor(w, TimeCursor);
	 if (loadfile(filename, mode))
	 {
	    if (FileBox)
	       XtDestroyWidget(w);
	    set_filename(filename);
	    XmTextSetCursorPosition(text_canvas, 0);
            set_status_message("");
	 }
	 else
	 {
	    set_status_message("Sorry! Unable to open file");
	 }
	 unset_cursor(w);
      }
   }
   else
   {
      set_status_message("Attempting to insert file");
      set_cursor(w, TimeCursor);
      if (loadfile(filename, mode))
      {
	    if (FileBox)
	       XtDestroyWidget(w);
            set_status_message("");
      }
      else
	 set_status_message("Sorry! Unable to open file");
      unset_cursor(w);
   }
}


/* automatically determine the wordwrap type the a filename */
static void AutoWW(char *name)
{
   int    i;
   char  *p;

   for (i=0; i<extnum; i++)
   {
      p = strrchr(name, '.');		/* search for extension */
      if (!p)
         continue;

      ++p;				/* skip dot */
      if (!strcmp(p, extmode[i].ext))
      {
         WordWrap   = extmode[i].ww;	/* if extension found in table, set */
         AutoIndent = extmode[i].ai;	/* the corresponding wordwrap and */
         SetAutoIndentAndWordWrap();	/* auto indent style */
         return;
      }
   }
}




/* Save the contents of the text widget to file */
void  gui_saveAsCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   gui_createFileDialog(toplevel, SAVE_AS);
}


/* Save the contents of the text widget to file */
void  gui_saveCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   if ( (strcmp(FileName, "") == 0) ||
        (strcmp(FileName, "-") == 0) ||
        (strcmp(FileName, "Untitled.gloe") == 0) )
   {
      gui_createFileDialog(toplevel, SAVE_ALL);
   }
   else
   {
      if (gui_save(FileName, SAVE_ALL))
      {
#ifdef PERSISTENT_UNDO
         clear_undo_redo();
#else
	 reset_undo_redo();
#endif
      }
   }
}

void  gui_saveSelectionCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   gui_createFileDialog(toplevel, SAVE_SEL);
}

static void  gui_saveFileCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   XmFileSelectionBoxCallbackStruct *cbd =
		(XmFileSelectionBoxCallbackStruct *)call_data;
   FileMode mode = (FileMode)client_data;
   char    *filename = NULL;

   /* Get the filename from the file selection box */
   XmStringGetLtoR(cbd->value, CHARSET, &filename);

   if (gui_save(filename, (int)mode))
   {
      set_filename(filename);
      if (mode == SAVE_ALL)
      {
#ifdef PERSISITENT_UNDO
         clear_undo_redo();
#else
	 reset_undo_redo();
#endif
      }
      else if (mode == SAVE_AS)
         reset_undo_redo();

      if (FileBox)
	 XtDestroyWidget(w);
   }
}

static Boolean gui_save(char *filename, FileMode mode)
{
   Boolean  flag;

   set_status_message("Attempting to save file");
   set_cursor(toplevel, TimeCursor);
   set_cursor(text_canvas, TimeCursor);

   flag = savefile(filename, mode);
   if (flag)
      set_status_message("Ok file saved");
   else if (gui_error("Sorry! Unable to save file ") == HELP)
      gui_info("You may not have specified a valid file name.\n"
	       "Or you may not have write permission to that file.");

   set_cursor(toplevel, MainCursor);
   set_cursor(text_canvas, TextCursor);

   return(flag);
}

static Boolean savefile(char *filename, FileMode mode)
{
   Boolean  flag = FALSE;
   char    *textcontents;
   FILE    *filepointer;

   if (*filename)
   {
      if (AutoBackup && (access(filename, F_OK) == 0))
      {
         DString  backup;
         DStringCompose(&backup, filename, ".backup", NULL);
	 rename(filename, backup.str);
	 DStringFree(&backup);
      }
      if ((filepointer = fopen(filename, "w+")) != NULL)
      {
	 /* Get the contents of the text widget text_canvas. */
	 if (mode == SAVE_SEL)
            textcontents = XmTextGetSelection(text_canvas);
         else
	    textcontents = XmTextGetString(text_canvas);
	 if (textcontents != NULL)
	 {
	    int  len = strlen(textcontents);

	    /* Write out the file */
	    fwrite(textcontents, sizeof(char), len, filepointer);
	    if (len && (textcontents[len - 1] != '\n'))
	       fputc('\n', filepointer);
 	    XtFree(textcontents);

	    flag = TRUE;
	 }
	 fclose(filepointer);
	 sync();
      }
   }
   return(flag);
}


void  gui_newCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   char     answer;

   /* Check to see if text has changed first before wiping it out */
   if (changes_counter)
      answer = gui_query("Unsaved Text. Are you sure about this?");
   else
      answer = YES;

   if (answer == SAVE_THEN_DO)
      gui_saveCB(w, NULL, NULL);

   if ((answer == YES) || (answer == SAVE_THEN_DO))
   {
      /* Now blank out the text_canvas and the file name fields */
      set_filename(Xstrdup("Untitled.gloe"));
      XmTextSetString(text_canvas, "");
      reset_undo_redo();
      SelectedOnly = False;
   }
}


void  gui_printCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   int      result;
   char    *printfile = tmpnam(NULL);

   set_status_message("Ok attempting to print file");
   if (savefile(printfile, SAVE_ALL))
   {
      DString  printcommand;
      DStringCompose(&printcommand, DefaultPrinter, " ", printfile, NULL);
      result = system(printcommand.str);
      DStringFree(&printcommand);
      if (result == 0)
	 set_status_message("Print succesfull");
      else
	 gui_info("Sorry unable to print file\n"
		  "Maybe you have specified an invalid printer");
      unlink(printfile);
   }
   else
   {
      if (gui_error("Sorry unable to print file ") == HELP)
	 gui_info("Sorry unable to write temporary file needed for printing\n"
		  "I need write permission for the directory '/tmp' to do this");
   }
}

void  gui_printSelectionCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   int      result;
   char    *printfile = tmpnam(NULL);

   set_status_message("Ok attempting to print file");

   if (savefile(printfile, SAVE_SEL))
   {
      DString  printcommand;
      DStringCompose(&printcommand, DefaultPrinter, " ", printfile);
      result = system(printcommand.str);
      DStringFree(&printcommand);
      if (result == 0)
	 set_status_message("Print succesfull");
      else
	 gui_info("Sorry unable to print file\n"
		  "Maybe you have specified an invalid printer");
      unlink(printfile);
   }
   else
   {
      if (gui_error("Sorry unable to save file ") == HELP)
	 gui_info("Sorry unable to write temporary file needed for printing\n"
		  "I need write permission for the directory '/tmp' to do this");
   }
}


void  gui_quitCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   char     answer;

   /* Check to see if text has changed first before wiping it out */
   if (changes_counter)
   {
      answer = gui_query("Unsaved text. Are you sure you wish to Quit?");

      if (answer == SAVE_THEN_DO)
	 gui_saveCB(w, NULL, NULL);

      if (answer == NO)
         return;
   }
   XtFree((char*)extmode);
   XtFree(FileName);
   exit(0);
}
